"""Test the SQLite store."""

import datetime
import sqlite3
from pathlib import Path
from typing import Generator

import pytest

from acoupi import components, data


@pytest.fixture(scope="function")
def sqlite_store(
    tmp_path: Path,
) -> Generator[components.SqliteStore, None, None]:
    """Create a store."""
    db_path = tmp_path / "test.db"
    yield components.SqliteStore(db_path)
    db_path.unlink()


def test_sqlite_store_creates_a_database_file(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the store creates a database file."""
    db_path = sqlite_store.db_path
    assert db_path.exists()


def test_database_has_correct_tables(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the database has the correct tables."""
    expected_tables = {
        "recording",
        "deployment",
        "detection",
    }
    db_path = sqlite_store.db_path

    # Check that the tables are there
    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        actual_tables = set(row[0] for row in cursor.fetchall())
        assert expected_tables.issubset(actual_tables)


def test_recording_table_has_correct_columns(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the recording table has the correct columns."""
    expected_columns = {
        "id",
        "path",
        "duration_s",
        "samplerate_hz",
        "audio_channels",
        "datetime",
        "deployment_id",
    }
    db_path = sqlite_store.db_path

    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("PRAGMA table_info(recording);")
        actual_columns = set(row[1] for row in cursor.fetchall())
        assert expected_columns == actual_columns


def test_deployment_table_has_correct_columns(sqlite_store) -> None:
    """Test that the deployment table has the correct columns."""
    expected_columns = {
        "id",
        "started_on",
        "ended_on",
        "latitude",
        "longitude",
        "name",
    }
    db_path = sqlite_store.db_path

    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("PRAGMA table_info(deployment);")
        actual_columns = set(row[1] for row in cursor.fetchall())
        assert expected_columns == actual_columns


def test_can_instantiate_multiple_stores_with_the_same_sqlite_file(
    tmp_path: Path,
) -> None:
    """Test that we can instantiate multiple stores to the same sqlite file."""
    # Arrange
    db_path = tmp_path / "test.db"

    # Act
    store1 = components.SqliteStore(db_path)
    store2 = components.SqliteStore(db_path)

    # Assert
    assert store1 is not None
    assert store2 is not None


def test_can_store_a_deployment(sqlite_store: components.SqliteStore) -> None:
    """Test that we can store a deployment."""
    # Arrange
    now = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=now,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )

    # Act
    sqlite_store.store_deployment(deployment)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM deployment;")
        row = cursor.fetchone()
        assert row["started_on"] == now.isoformat(sep=" ")
        assert row["latitude"] == 1.0
        assert row["longitude"] == 2.0


def test_get_current_deployment_returns_new_if_no_deployment_exists(
    sqlite_store: components.SqliteStore,
    patched_now,
) -> None:
    """Test that get_current_deployment fails if no deployment exists."""
    # Arrange
    # Patch the current time
    now = patched_now()

    # Make sure there are no deployments
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("DELETE FROM deployment;")

    # Act
    # Try to get the current deployment, this should create a new one
    deployment = sqlite_store.get_current_deployment()

    # Assert
    assert deployment.latitude is None
    assert deployment.longitude is None
    assert deployment.started_on == now


def test_get_current_deployment_gets_latest_deployment(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that get_current_deployment gets the latest deployment."""
    # Arrange
    start_1 = datetime.datetime.now()
    start_2 = datetime.datetime.now() - datetime.timedelta(days=10)
    deployment1 = data.Deployment(
        started_on=start_1,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    deployment2 = data.Deployment(
        started_on=start_2,
        latitude=3.0,
        longitude=4.0,
        name="test_device",
    )
    sqlite_store.store_deployment(deployment1)
    sqlite_store.store_deployment(deployment2)

    # Act
    deployment = sqlite_store.get_current_deployment()

    # Assert
    assert deployment == deployment1


def test_deployment_id_in_db_is_same_as_in_python(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the deployment_id in the db is the same as in python."""
    # Arrange
    now = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=now,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )

    # Act
    sqlite_store.store_deployment(deployment)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM deployment;")
        row = cursor.fetchone()
        assert row["id"] == deployment.id.bytes


def test_recordings_can_be_registered(
    sqlite_store: components.SqliteStore,
    deployment: data.Deployment,
) -> None:
    """Test that recordings can be registered."""
    # Arrange
    now = datetime.datetime.now()
    recording = data.Recording(
        deployment=deployment,
        path=Path("test/path"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
        audio_channels=2,
    )

    # Act
    sqlite_store.store_recording(recording)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM recording;")
        row = cursor.fetchone()
        assert row["path"] == "test/path"
        assert row["duration_s"] == 10.0
        assert row["audio_channels"] == 2
        assert row["samplerate_hz"] == 44100
        assert row["datetime"] == now.isoformat(sep=" ")
        assert row["id"] == recording.id.bytes


def test_recording_can_be_registered_with_custom_deployment(
    sqlite_store: components.SqliteStore,
):
    """Test that recordings can be registered with a custom deployment."""
    # Arrange
    now = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=now,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    sqlite_store.store_deployment(deployment)
    recording = data.Recording(
        path=Path("test/path"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
        deployment=deployment,
    )

    # Act
    sqlite_store.store_recording(recording)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM recording;")
        row = cursor.fetchone()
        assert row["path"] == "test/path"
        assert row["duration_s"] == 10.0
        assert row["samplerate_hz"] == 44100
        assert row["datetime"] == now.isoformat(sep=" ")
        assert row["id"] == recording.id.bytes
        assert row["deployment_id"] == deployment.id.bytes


def test_get_all_recordings(
    sqlite_store: components.SqliteStore,
    deployment: data.Deployment,
):
    """Test that all recordings can be retrieved."""
    # Arrange

    # Create two recordings
    now = datetime.datetime.now()
    recording1 = data.Recording(
        deployment=deployment,
        path=Path("test/path1"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
    )
    recording2 = data.Recording(
        deployment=deployment,
        path=Path("test/path2"),
        duration=10.0,
        samplerate=44100,
        created_on=now + datetime.timedelta(seconds=10),
    )

    # Add them to the database
    sqlite_store.store_recording(recording1)
    sqlite_store.store_recording(recording2)

    # Act
    recordings = sqlite_store.get_recordings([recording1.id, recording2.id])

    # Assert
    assert len(recordings) == 2
    assert recordings[1][0] == recording1
    assert recordings[0][0] == recording2


def test_get_some_recordings(
    sqlite_store: components.SqliteStore,
    deployment: data.Deployment,
):
    """Test that recordings can be retrieved with an include filter."""
    # Arrange

    # Create two recordings
    now = datetime.datetime.now()
    recording1 = data.Recording(
        deployment=deployment,
        path=Path("test/path1"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
    )
    recording2 = data.Recording(
        deployment=deployment,
        path=Path("test/path2"),
        duration=10.0,
        samplerate=44100,
        created_on=now + datetime.timedelta(seconds=10),
    )

    # Add them to the database
    sqlite_store.store_recording(recording1)
    sqlite_store.store_recording(recording2)

    # Act
    recordings = sqlite_store.get_recordings([recording1.id])

    # Assert
    assert len(recordings) == 1
    assert recordings[0][0] == recording1


def test_can_store_model_outputs(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    sqlite_store.store_model_output(model_output)
    db_path = sqlite_store.db_path

    retrieved = sqlite_store.get_recordings([model_output.recording.id])
    assert len(retrieved) == 1
    assert retrieved[0][1][0] == model_output

    # Check that the detections were stored
    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT id FROM model_output;")
        assert len(cursor.fetchall()) == 1

        cursor.execute("SELECT id FROM detection;")
        assert len(cursor.fetchall()) == 1


def test_can_update_deployment_info(
    sqlite_store: components.SqliteStore,
):
    start = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=start,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    sqlite_store.store_deployment(deployment)

    current_deployment = sqlite_store.get_current_deployment()
    assert deployment == current_deployment

    assert current_deployment.ended_on is None

    updated_deployment = deployment.model_copy(
        update=dict(ended_on=start + datetime.timedelta(days=2))
    )
    assert updated_deployment.ended_on is not None
    sqlite_store.update_deployment(updated_deployment)

    current_deployment = sqlite_store.get_current_deployment()
    assert updated_deployment.model_dump() == current_deployment.model_dump()
    assert current_deployment.ended_on is not None


def test_update_deployment_fails_if_corresponding_deployment_does_not_exist(
    sqlite_store: components.SqliteStore,
):
    start = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=start,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    with pytest.raises(ValueError):
        sqlite_store.update_deployment(deployment)



"""Acoupi devices module.

This module provides functions to retrieve information about the device
on which Acoupi is running. This includes identifying the device type
(currently supports Raspberry Pi), and obtaining unique identifiers
like serial numbers and hostnames.
"""

import uuid

from acoupi.devices.rpi import get_rpi_host_name, get_rpi_serial_number, is_rpi


def get_device_id() -> str:
    """Get a unique identifier for the current device.

    This function returns a device-specific identifier.

    - If the device is a Raspberry Pi, the serial number is returned.
    - Otherwise, the MAC address is used as the identifier,
      retrieved via `uuid.getnode()`.

    Returns
    -------
    str
        The device ID.

    Notes
    -----
        For potential limitations of using MAC addresses for identification,
        refer to the [`uuid.getnode()`][uuid.getnode] documentation.
    """
    if is_rpi():
        return get_rpi_serial_number()

    return str(uuid.getnode())


__all__ = [
    "get_rpi_serial_number",
    "get_rpi_host_name",
    "get_device_id",
    "is_rpi",
]



"""Module defining the SqliteStore class."""

import datetime
import json
import warnings
from pathlib import Path
from typing import List, Optional, Tuple
from uuid import UUID

from pony import orm

from . import types as db_types
from .database import create_base_models
from acoupi import data
from acoupi.components import types

db_session = orm.db_session()


class SqliteStore(types.Store):
    """Sqlite store implementation.

    The store is used to store the recordings, detections and deployments
    locally. The data is stored in a sqlite database file in the given path.

    Under the hood, the store uses the Pony ORM to interact with the database.
    The database schema is defined in the database module and contains the
    following tables:

    - Deployment: Contains the deployment information. Each deployment is
      associated with a device, and has a start datetime. The deployment can
      also have a latitude and longitude associated with it.

    - Recording: Contains the recording information. Each recording is
      associated with a deployment, and has a datetime, duration, samplerate
      and number of                 audio_channels.

    - PredictedTag: Contains the predicted tag information. Each predicted tag
      has a key, value and score.

    - Detection: Contains the detection information. Each detection consists
      of a location, score and a list of predicted tags.

    - ModelOutput: Contains the model output information. Each model output
      has the model name, the list of predicted tags at the recording level,
      and a list of detections.

    The store is thread-safe, and can be used from multiple threads
    simultaneously.

    Notes
    -----
    The ID of the deployment, recording, model output and detection is a UUID
    field. Note that sqlite stores UUIDs as a BLOB, so when querying the
    database with SQL, you should use the hex function to convert the UUID to a string.

    Example queries:

    .. code-block:: sql

        -- Find a specific deployment by UUID
        SELECT * FROM Deployment WHERE hex(id) = '00000000000000000000000000000000';

        -- Get all deployment UUIDs as strings
        SELECT hex(id) FROM Deployment;
    """

    db_path: Path
    """Path to the database file."""

    database: orm.Database
    """The Pony ORM database object."""

    models: db_types.BaseModels
    """The Pony ORM models."""

    def __init__(self, db_path: Path) -> None:
        """Initialise the Sqlite Store.

        Will create a database file at the given path if it does not exist.

        Args:
            db_path: Path to the database file. Can be set to :memory: to use
                an in-memory database.
        """
        self.db_path = db_path
        self.database = orm.Database()
        self.models = create_base_models(self.database)
        self.database.bind(
            provider="sqlite",
            filename=str(db_path),
            create_db=True,
        )
        self.database.generate_mapping(create_tables=True)

    @db_session
    def get_current_deployment(self) -> data.Deployment:
        """Get the current deployment.

        The current deployment is the one with the latest started_on datetime.

        If no deployment is found, a new deployment will be registered with the
        current datetime, and the latitude and longitude set to None.

        Returns
        -------
            The current deployment
        """
        deployment = self._get_current_deployment()
        return data.Deployment(
            id=deployment.id,
            name=deployment.name,
            latitude=deployment.latitude,
            longitude=deployment.longitude,
            started_on=deployment.started_on,
            ended_on=deployment.ended_on,
        )

    def store_deployment(self, deployment: data.Deployment) -> None:
        """Store the deployment locally.

        Args:
            deployment: The deployment to store
        """
        self._get_or_create_deployment(deployment)

    @db_session
    def update_deployment(self, deployment: data.Deployment) -> None:
        db_deployment = self._get_deployment_by_id(deployment.id)
        db_deployment.name = deployment.name
        db_deployment.latitude = deployment.latitude
        db_deployment.longitude = deployment.longitude

        if deployment.ended_on is not None:
            db_deployment.ended_on = deployment.ended_on

    @db_session
    def store_recording(self, recording: data.Recording) -> None:
        """Store the recording locally.

        If the deployment is not provided, the current deployment will be used.

        Args:
            recording: The recording to store
        """
        self._get_or_create_recording(recording)

    @db_session
    def store_model_output(self, model_output: data.ModelOutput) -> None:
        """Store the model output locally."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=orm.PonyRuntimeWarning)
            db_recording = self._get_or_create_recording(
                model_output.recording
            )

        db_model_output = self.models.ModelOutput(
            id=model_output.id,
            model_name=model_output.name_model,
            recording=db_recording,
            created_on=model_output.created_on,
        )

        for tag in model_output.tags:
            db_model_output.tags.create(
                key=tag.tag.key,
                value=tag.tag.value,
                confidence_score=tag.confidence_score,
            )

        for detection in model_output.detections:
            location = (
                ""
                if detection.location is None
                else detection.location.model_dump_json()
            )
            db_detection = db_model_output.detections.create(
                id=detection.id,
                location=location,
                detection_score=detection.detection_score,
            )

            for tag in detection.tags:
                db_detection.tags.create(
                    key=tag.tag.key,
                    value=tag.tag.value,
                    confidence_score=tag.confidence_score,
                )

    @db_session
    def get_recordings_by_path(
        self,
        paths: List[Path],
    ) -> List[Tuple[data.Recording, List[data.ModelOutput]]]:
        """Get a list of recordings from the store by their paths.

        Args:
            paths: The paths of the recordings to get.

        Returns
        -------
            List of tuples of the recording and the corresponding model outputs.
        """
        paths_str = [str(path) for path in paths]
        db_recordings = (
            orm.select(r for r in self.models.Recording if r.path in paths_str)
            .order_by(orm.desc(self.models.Recording.datetime))
            .prefetch(
                self.models.Recording.deployment,
                self.models.Recording.model_outputs,
                self.models.ModelOutput.tags,
                self.models.ModelOutput.detections,
                self.models.Detection.tags,
            )
        )
        return _to_recordings_and_outputs(db_recordings)

    @db_session
    def get_recordings(
        self,
        ids: List[UUID],
    ) -> List[Tuple[data.Recording, List[data.ModelOutput]]]:
        """Get a list recordings from the store by their ids.

        Each recording is returned with the full list of model outputs
        registered.

        Args:
            ids: The ids of the recordings to get.

        Returns
        -------
            A list of tuples of the recording and the model outputs.
        """
        db_recordings = (
            orm.select(r for r in self.models.Recording if r.id in ids)
            .order_by(orm.desc(self.models.Recording.datetime))
            .prefetch(
                self.models.Recording.deployment,
                self.models.Recording.model_outputs,
                self.models.ModelOutput.tags,
                self.models.ModelOutput.detections,
                self.models.Detection.tags,
            )
        )

        return _to_recordings_and_outputs(db_recordings)

    @db_session
    def get_model_outputs(
        self,
        after: Optional[datetime.datetime] = None,
        before: Optional[datetime.datetime] = None,
        ids: Optional[List[UUID]] = None,
        recording_ids: Optional[List[UUID]] = None,
        model_names: Optional[List[str]] = None,
        detection_ids: Optional[List[UUID]] = None,
        limit: Optional[int] = None,
    ) -> List[data.ModelOutput]:
        """Get a list of model outputs from the store by their id.

        Args:
            start_time: The time to start the search from.
            end_time: The time to end the search at.

        Returns
        -------
            List of model_outputs matching the created_on datetime.
        """
        query = orm.select(mo for mo in self.models.ModelOutput)

        if after is not None:
            query = query.filter(lambda mo: mo.created_on >= after)

        if before is not None:
            query = query.filter(lambda mo: mo.created_on <= before)

        if ids is not None:
            query = query.filter(lambda mo: mo.id in ids)

        if recording_ids is not None:
            query = query.filter(lambda mo: mo.recording.id in recording_ids)

        if model_names is not None:
            query = query.filter(lambda mo: mo.model_name in model_names)

        if detection_ids is not None:
            query = query.filter(
                lambda mo: orm.exists(
                    d for d in mo.detections if d.id in detection_ids
                )
            )

        query = query.order_by(
            orm.desc(self.models.ModelOutput.created_on)
        ).prefetch(
            self.models.ModelOutput.tags,
            self.models.ModelOutput.detections,
            self.models.Detection.tags,
        )

        if limit is not None:
            query = query[:limit]

        return [_to_model_output(db_model_output) for db_model_output in query]

    @db_session
    def get_detections(
        self,
        ids: Optional[List[UUID]] = None,
        model_output_ids: Optional[List[UUID]] = None,
        score_gt: Optional[float] = None,
        score_lt: Optional[float] = None,
        model_names: Optional[List[str]] = None,
        after: Optional[datetime.datetime] = None,
        before: Optional[datetime.datetime] = None,
    ) -> List[data.Detection]:
        """Get a list of detections from the store based on their model_output ids."""
        query = orm.select(d for d in self.models.Detection)

        if ids:
            query = query.filter(lambda d: d.id in ids)

        if model_output_ids:
            query = query.filter(
                lambda d: d.model_output.id in model_output_ids
            )

        if score_gt is not None:
            query = query.filter(lambda d: d.detection_score > score_gt)

        if score_lt is not None:
            query = query.filter(lambda d: d.detection_score < score_lt)

        if after is not None:
            query = query.filter(lambda d: d.model_output.created_on >= after)

        if before is not None:
            query = query.filter(lambda d: d.model_output.created_on <= before)

        if model_names:
            query = query.filter(
                lambda d: d.model_output.model_name in model_names
            )

        query = query.prefetch(self.models.Detection.tags)

        return [_to_detection(db_detection) for db_detection in query]

    @db_session
    def get_predicted_tags(
        self,
        detection_ids: Optional[List[UUID]] = None,
        after: Optional[datetime.datetime] = None,
        before: Optional[datetime.datetime] = None,
        score_gt: Optional[float] = None,
        score_lt: Optional[float] = None,
        keys: Optional[List[str]] = None,
        values: Optional[List[str]] = None,
    ) -> List[data.PredictedTag]:
        """Get a list of predicted tags from the store based on their detection ids."""
        query = orm.select(t for t in self.models.PredictedTag)

        if detection_ids:
            query = query.filter(
                lambda t: t.detection is not None
                and t.detection.id in detection_ids
            )

        if after is not None:
            query = query.filter(
                lambda t: t.detection is not None
                and t.detection.model_output.created_on >= after
            )

        if before is not None:
            query = query.filter(
                lambda t: t.detection is not None
                and t.detection.model_output.created_on <= before
            )

        if score_gt is not None:
            query = query.filter(lambda t: t.confidence_score > score_gt)

        if score_lt is not None:
            query = query.filter(lambda t: t.confidence_score < score_lt)

        if keys:
            query = query.filter(lambda t: t.key in keys)

        if values:
            query = query.filter(lambda t: t.value in values)

        return [_to_predictedtag(db_tag) for db_tag in query]

    @db_session
    def update_recording_path(
        self,
        recording: data.Recording,
        path: Path,
    ) -> data.Recording:
        """Update the path of a recording.

        Args:
            recording: The recording to update.
            path: The new path.
        """
        db_recording = self._get_or_create_recording(recording)
        db_recording.path = str(path)
        recording.path = path
        return recording

    @db_session
    def _get_current_deployment(self) -> db_types.Deployment:
        """Get the current deployment in the database."""
        db_deployment = (
            orm.select(d for d in self.models.Deployment)
            .order_by(orm.desc(self.models.Deployment.started_on))
            .first()
        )

        if db_deployment is None:
            now = datetime.datetime.now()
            name = f'Deployment {now.strftime("%Y-%m-%d %H:%M:%S")}'
            db_deployment = self.models.Deployment(
                started_on=now,
                name=name,
                latitude=None,
                longitude=None,
            )

        return db_deployment

    @db_session
    def _create_recording(
        self,
        recording: data.Recording,
    ) -> db_types.Recording:
        """Create a recording."""
        deployment_db = self._get_or_create_deployment(recording.deployment)
        db_recording = self.models.Recording(
            id=recording.id,
            path=str(recording.path),
            duration_s=recording.duration,
            samplerate_hz=recording.samplerate,
            audio_channels=recording.audio_channels,
            datetime=recording.created_on,
            deployment=deployment_db,
        )
        return db_recording

    def _get_or_create_recording(
        self,
        recording: data.Recording,
    ) -> db_types.Recording:
        """Get or create a recording."""
        try:
            return self._get_recording_by_id(recording.id)
        except ValueError:
            return self._create_recording(recording)

    @db_session
    def _get_deployment_by_started_on(
        self, started_on: datetime.datetime
    ) -> db_types.Deployment:
        """Get the deployment by the started_on datetime."""
        deployment: Optional[db_types.Deployment] = self.models.Deployment.get(
            started_on=started_on
        )

        if deployment is None:
            raise ValueError("No deployment found")

        return deployment

    @db_session
    def _create_deployment(
        self,
        deployment: data.Deployment,
    ) -> db_types.Deployment:
        """Create a deployment."""
        db_deployment = self.models.Deployment(
            id=deployment.id,
            started_on=deployment.started_on,
            name=deployment.name,
            latitude=deployment.latitude,
            longitude=deployment.longitude,
        )
        return db_deployment

    def _get_or_create_deployment(
        self, deployment: data.Deployment
    ) -> db_types.Deployment:
        """Get or create a deployment."""
        try:
            db_deployment = self._get_deployment_by_id(deployment.id)
        except ValueError:
            db_deployment = self._create_deployment(deployment)

        return db_deployment

    @db_session
    def _get_deployment_by_id(self, id: UUID) -> db_types.Deployment:
        """Get the deployment by the id."""
        deployment: Optional[db_types.Deployment] = self.models.Deployment.get(
            id=id
        )

        if deployment is None:
            raise ValueError("No deployment found")

        return deployment

    @db_session
    def _get_recording_by_id(self, id: UUID) -> db_types.Recording:
        """Get the recording by the id."""
        recording: Optional[db_types.Recording] = self.models.Recording.get(
            id=id
        )

        if recording is None:
            raise ValueError("No recording found")

        return recording


def _to_deployment(db_deployment: db_types.Deployment) -> data.Deployment:
    return data.Deployment(
        id=db_deployment.id,
        name=db_deployment.name,
        latitude=db_deployment.latitude,
        longitude=db_deployment.longitude,
        started_on=db_deployment.started_on,
    )


def _to_recording(db_recording: db_types.Recording) -> data.Recording:
    deployment = _to_deployment(db_recording.deployment)
    return data.Recording(
        id=db_recording.id,
        deployment=deployment,
        created_on=db_recording.datetime,
        duration=db_recording.duration_s,
        samplerate=db_recording.samplerate_hz,
        audio_channels=db_recording.audio_channels,
        path=None if db_recording.path is None else Path(db_recording.path),
    )


def _to_predictedtag(db_tag: db_types.PredictedTag) -> data.PredictedTag:
    return data.PredictedTag(
        tag=data.Tag(
            key=db_tag.key,
            value=db_tag.value,
        ),
        confidence_score=db_tag.confidence_score,
    )


def _to_detection(db_detection: db_types.Detection) -> data.Detection:
    location = (
        None
        if db_detection.location == ""
        else json.loads(str(db_detection.location))
    )
    return data.Detection(
        id=db_detection.id,
        location=location,
        detection_score=db_detection.detection_score,
        tags=[
            data.PredictedTag(
                tag=data.Tag(
                    key=db_tag.key,
                    value=db_tag.value,
                ),
                confidence_score=db_tag.confidence_score,
            )
            for db_tag in db_detection.tags
        ],
    )


def _to_model_output(
    db_model_output: db_types.ModelOutput,
) -> data.ModelOutput:
    recording = _to_recording(db_model_output.recording)
    return data.ModelOutput(
        id=db_model_output.id,
        name_model=db_model_output.model_name,
        recording=recording,
        created_on=db_model_output.created_on,
        tags=[
            data.PredictedTag(
                tag=data.Tag(
                    key=db_tag.key,
                    value=db_tag.value,
                ),
                confidence_score=db_tag.confidence_score,
            )
            for db_tag in db_model_output.tags
        ],
        detections=[
            _to_detection(db_detection)
            for db_detection in db_model_output.detections
        ],
    )


def _to_recordings_and_outputs(
    db_recordings,
) -> List[Tuple[data.Recording, List[data.ModelOutput]]]:
    ret = []
    for db_recording in db_recordings:
        recording = _to_recording(db_recording)

        outputs = []
        for db_model_output in db_recording.model_outputs:
            model_output = _to_model_output(db_model_output)
            outputs.append(model_output)

        ret.append((recording, outputs))

    return ret



"""Data objects for acoupi System."""

import datetime
from enum import IntEnum
from pathlib import Path
from typing import List, Optional, Tuple
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator

__all__ = [
    "TimeInterval",
    "Deployment",
    "Recording",
    "PredictedTag",
    "Detection",
    "ModelOutput",
    "Message",
    "ResponseStatus",
    "Response",
    "BoundingBox",
]


class TimeInterval(BaseModel):
    """An interval of time between two times of day."""

    start: datetime.time
    """Start time of the interval."""

    end: datetime.time
    """End time of the interval."""


class Deployment(BaseModel):
    """A Deployment captures information about the device deployment.

    This includes the latitude, longitude, and deployment start.
    """

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the deployment."""

    name: str
    """User provided name of the deployment."""

    latitude: Optional[float] = None
    """The latitude of the site where the device is deployed."""

    longitude: Optional[float] = None
    """The longitude of the site where the device is deployed."""

    started_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime when the device was deployed."""

    ended_on: Optional[datetime.datetime] = None
    """The datetime when the deployment ended."""

    @field_validator("latitude")
    def validate_latitude(cls, value):
        """Validate that the latitude are within range."""
        if value is not None and (value < -90 or value > 90):
            raise ValueError("latitude must be between -90 and 90")
        return value

    @field_validator("longitude")
    def validate_longitude(cls, value):
        """Validate that the longitude are within range."""
        if value is not None and (value < -180 or value > 180):
            raise ValueError("longitude must be between -180 and 180")
        return value


class Recording(BaseModel):
    """A Recording is a single audio file recorded from the microphone."""

    created_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now,
        repr=False,
    )
    """The datetime when the recording was made"""

    duration: float = Field(
        repr=False,
    )
    """The duration of the recording in seconds"""

    samplerate: int = Field(repr=False)
    """The samplerate of the recording in Hz"""

    deployment: Deployment = Field(repr=False)
    """The deployment that the recording belongs to"""

    path: Optional[Path] = Field(None, repr=True)
    """The path to the audio file in the local filesystem"""

    audio_channels: Optional[int] = Field(default=1, repr=False)
    """The number of audio_channels in the recording."""

    chunksize: Optional[int] = Field(default=4096, repr=False)
    """The chunksize of the audio file in bytes. Defaults to 4096."""

    id: UUID = Field(default_factory=uuid4, repr=True)
    """The unique ID of the recording"""

    @field_validator("duration")
    def validate_duration(cls, value):
        """Validate that the duration is greater than 0."""
        if value <= 0:
            raise ValueError("duration must be greater than 0")
        return value

    @field_validator("samplerate")
    def validate_samplerate(cls, value):
        """Validate that the samplerate is greater than 0."""
        if value <= 0:
            raise ValueError("samplerate must be greater than 0")
        return value

    @field_validator("audio_channels")
    def validate_audio_channels(cls, value):
        """Validate that the number of audio_channels is greater than 1."""
        if value <= 0:
            raise ValueError("audio_channels must be 1 or greater")
        return value


class Tag(BaseModel):
    """A Tag is a label for a recording."""

    key: str
    """The key of the tag."""

    value: str
    """The value of the tag."""

    @field_validator("key")
    def validate_key(cls, value):
        """Validate that the key is not empty."""
        if not value:
            raise ValueError("key cannot be empty")
        return value

    @field_validator("value")
    def validate_value(cls, value):
        """Validate that the value is not empty."""
        if not value:
            raise ValueError("value cannot be empty")
        return value


class PredictedTag(BaseModel):
    """A PredictedTag is a label predicted by a model.

    It consists of a key, a value and a score.
    """

    tag: Tag
    """The tag predicted by the model."""

    confidence_score: float = 1
    """The confidence score of the predicted tag."""

    @field_validator("confidence_score")
    def validate_score(cls, value):
        """Validate that the score is between 0 and 1."""
        if value < 0 or value > 1:
            raise ValueError("score must be between 0 and 1")
        return value


class BoundingBox(BaseModel):
    """BoundingBox to locate a sound event in time and frequency.

    All time values are in seconds and all frequency values are in Hz.
    """

    type: str = "BoundingBox"

    coordinates: Tuple[float, float, float, float]

    @classmethod
    def from_coordinates(
        cls,
        start_time: float,
        low_freq: float,
        end_time: float,
        high_freq: float,
    ):
        """Create a BoundingBox from coordinates."""
        return cls(
            coordinates=(start_time, low_freq, end_time, high_freq),
        )

    @field_validator("coordinates")
    def validate_coordinates(cls, value):
        """Validate that the coordinates are within range."""
        start_time, low_freq, end_time, high_freq = value

        if end_time <= start_time:
            raise ValueError("end time must be greater than start time")

        if low_freq >= high_freq:
            raise ValueError(
                "high frequency must be greater than low frequency"
            )

        return value


class Detection(BaseModel):
    """A Detection is a single prediction from a model."""

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the detection"""

    location: Optional[BoundingBox] = None
    """The location of the detection in the recording."""

    detection_score: float = 1
    """The score of the detection."""

    tags: List[PredictedTag] = Field(default_factory=list)
    """The tags predicted by the model for the detection."""

    @field_validator("detection_score")
    def validate_score(cls, value):
        """Validate that the score is between 0 and 1."""
        if value < 0 or value > 1:
            raise ValueError("score must be between 0 and 1")
        return value

    @field_validator("tags")
    def sort_tags(cls, value):
        """Sort tags."""
        return sorted(
            value,
            key=lambda x: (
                x.confidence_score,
                x.tag.key,
                x.tag.value,
            ),
            reverse=True,
        )


class ModelOutput(BaseModel):
    """The output of a model."""

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the model output."""

    name_model: str
    """The name of the model that produced the output."""

    recording: Recording
    """The recording that was used as input to the model."""

    tags: List[PredictedTag] = Field(default_factory=list)
    """The tags predicted by the model at the recording level."""

    detections: List[Detection] = Field(default_factory=list)
    """List of predicted sound events in the recording."""

    created_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime when the model output was created."""

    @field_validator("tags")
    def sort_tags(cls, value):
        """Sort tags."""
        return sorted(
            value,
            key=lambda x: (
                x.confidence_score,
                x.tag.key,
                x.tag.value,
            ),
            reverse=True,
        )

    @field_validator("detections")
    def sort_detections(cls, value):
        """Sort detections by ID."""
        return sorted(
            value,
            key=lambda x: x.id,
        )


class Message(BaseModel):
    """The message to be sent to the remote server."""

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the message."""

    content: str
    """The message to be sent. Usually a JSON string."""

    created_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime when the message was created."""


class ResponseStatus(IntEnum):
    """The status of a message."""

    SUCCESS = 0
    """The message was received successfully."""

    FAILED = 1
    """The message failed to send."""

    ERROR = 2
    """The message was sent, but there was an error."""

    TIMEOUT = 3
    """The message timed out."""


class Response(BaseModel):
    """The response from sending a message."""

    status: ResponseStatus
    """The status of the message."""

    message: Message
    """The message that was sent."""

    content: Optional[str] = None
    """The content of the response."""

    received_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime the message was received."""



"""Test suite for the file management task generator."""

import datetime
from pathlib import Path
from typing import List, Optional

import pytest

from acoupi import data
from acoupi.components import SqliteStore
from acoupi.components.types import (
    RecordingSavingFilter,
    RecordingSavingManager,
)
from acoupi.tasks import generate_file_management_task


class DummyRecordingManager(RecordingSavingManager):
    def __init__(self, path: Path):
        self.path = path
        if not self.path.exists():
            self.path.mkdir(parents=True)

    def save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> Path:
        if recording.path is None:
            raise ValueError("Recording has no path")

        return self.path / recording.path.name


class DummyFileFilter(RecordingSavingFilter):
    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        return False


@pytest.fixture
def temp_audio_dir(tmp_path: Path) -> Path:
    path = tmp_path / "tmp_audio"
    if not path.exists():
        path.mkdir()
    return path


@pytest.fixture
def recording(
    temp_audio_dir: Path, deployment: data.Deployment
) -> data.Recording:
    recording_path = temp_audio_dir / "test.wav"
    recording_path.touch()
    return data.Recording(
        path=recording_path,
        duration=1,
        samplerate=256000,
        audio_channels=1,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )


@pytest.fixture
def model_output(recording: data.Recording) -> data.ModelOutput:
    return data.ModelOutput(
        name_model="test_model",
        recording=recording,
        tags=[
            data.PredictedTag(
                tag=data.Tag(key="test", value="value1"),
                confidence_score=0.8,
            ),
            data.PredictedTag(
                tag=data.Tag(key="test", value="value2"),
                confidence_score=0.8,
            ),
        ],
        detections=[
            data.Detection(
                location=data.BoundingBox(coordinates=(1, 1000, 2, 2000)),
                detection_score=0.6,
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(key="test2", value="value3"),
                        confidence_score=0.3,
                    ),
                    data.PredictedTag(
                        tag=data.Tag(key="test", value="value1"),
                        confidence_score=0.2,
                    ),
                ],
            )
        ],
    )


@pytest.fixture
def store(tmp_path: Path, deployment: data.Deployment) -> SqliteStore:
    store = SqliteStore(tmp_path / "test.db")
    store.store_deployment(deployment)
    return store


@pytest.fixture
def target_dir(tmp_path: Path) -> Path:
    return tmp_path / "target_dir"


@pytest.fixture
def dummy_manager(target_dir: Path) -> DummyRecordingManager:
    return DummyRecordingManager(target_dir)


def test_file_management_task_keeps_unprocessed_recordings(
    recording: data.Recording,
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        required_models=["test_model"],
    )

    task()

    assert recording.path is not None
    assert recording.path.exists()


def test_file_management_moves_files_if_no_required_model(
    target_dir: Path,
    recording: data.Recording,
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
    )

    task()

    recording, _ = store.get_recordings([recording.id])[0]
    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(target_dir)


def test_file_management_ignores_files_without_path(
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    deployment: data.Deployment,
    store: SqliteStore,
):
    recording = data.Recording(
        path=None,
        duration=1,
        samplerate=256000,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )
    store.store_recording(recording)
    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
    )

    task()


def test_file_management_deletes_files_that_do_not_pass_filters(
    target_dir: Path,
    recording: data.Recording,
    temp_audio_dir: Path,
    model_output: data.ModelOutput,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)
    store.store_model_output(model_output)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        file_filters=[DummyFileFilter()],
    )

    task()

    assert recording.path is not None
    assert not recording.path.exists()
    assert not (target_dir / recording.path.name).exists()
